@@ -23,20 +23,24 @@
 
 import asyncio
 import websockets
+from time import time
 from uuid import uuid4
 from json import loads, dumps
 from logging import info, debug
+from traceback import print_exc
 from inspect import signature, Parameter
 from concurrent.futures import ThreadPoolExecutor
 from typing import Awaitable, List, Union, Tuple, Any
 from websockets.exceptions import ConnectionClosedOK, ConnectionClosedError
 
 from .utils import Repr
-from .config import apiUrl, heartbeatInterval, topPublicRoomsInterval
 from .utils.convertors import Convertor
+from .ext import telemetry as telemetry_extension
+from .config import apiUrl, heartbeatInterval, topPublicRoomsInterval
 from .entities import Client, User, Room, UserPreview, Message, BaseUser, Context
 from .utils.parsers import parse_sentence_to_tokens as parse_message, parse_word_to_token as parse_word
-from .exceptions import NoConnectionException, InvalidAccessToken, InvalidSize, NotEnoughArguments, CommandNotFound, MemberNotFound
+from .exceptions import NoConnectionException, InvalidAccessToken, InvalidSize, NotEnoughArguments, CommandNotFound, MemberNotFound, CommandAlreadyDefined
+
 
 listeners = {}
 commands = {}
@@ -59,9 +63,15 @@ def event(func: Awaitable):
     return func
 
 
-def command(func: Awaitable = None, *, name: str = None):
+def command(func: Awaitable = None, *, name: str = None, cooldown: int = 0, aliases: List[str] = None):
     """
-    Create a new command for dogehouse.
+    Create a new DogeClient command, which will be handled by the dogehouse python library.
+
+    Args:
+        func (Awaitable, optional): Your function (gets supplied automatically when you use the decorator). Defaults to None.
+        name (str, optional): The call name for your command. Defaults to the function name.
+        cooldown (int, optional): The cooldown for the function usage per client. Defaults to 0.
+        aliases (List[str], optional): A list of aliases which should trigger your command. Defaults to None.
 
     Example:
         class Client(dogehouse.DogeClient):
@@ -73,7 +83,15 @@ def command(func: Awaitable = None, *, name: str = None):
             Client("token", "refresh_token").run()
     """
     def wrapper(func: Awaitable):
-        commands[str(name if name else func.__name__).lower()] = [func, False]
+        def save(name: str):
+            if name in commands:
+                raise CommandAlreadyDefined(f"Command `{name}` has already been defined by a name or alias!")
+            commands[name] = [func, False, int(cooldown)]
+        
+        save(str(name if name else func.__name__).lower())
+        if aliases:
+            any(save(alias) for alias in aliases)
+        
         return func
     return wrapper(func) if func else wrapper
 
@@ -81,7 +99,7 @@ def command(func: Awaitable = None, *, name: str = None):
 class DogeClient(Client):
     """Represents your Dogehouse client."""
 
-    def __init__(self, token: str, refresh_token: str, *, room: str = None, muted: bool = False, reconnect_voice: bool = False, prefix: Union[str, List[str]] = "!"):
+    def __init__(self, token: str, refresh_token: str, *, room: str = None, muted: bool = False, reconnect_voice: bool = False, prefix: Union[str, List[str]] = "!", telemetry: bool = False):
         """
         Initialize your Dogehouse client
 
@@ -91,7 +109,8 @@ class DogeClient(Client):
             room (int, optional): The room your client should join. Defaults to None.
             muted (bool, optional): Wether or not the client should be muted. Defaults to False.
             reconnect_voice (bool, optional): When the client disconnects from the voice server, should it try to reconnect. Defaults to False.
-            prefix (List of strings or a string): The bot prefix.
+            prefix (List of strings or a string, optional): The bot prefix.
+            telemetry (bool, optional): Whether or not you want to enable telemetry for your bot. Defaults to False.
         """
         super().__init__(None, room, [], prefix)
 
@@ -106,6 +125,11 @@ class DogeClient(Client):
         self.__commands = commands
         self.__waiting_for = {}
         self.__waiting_for_fetches = {}
+        self.__command_cooldown_cache = {}
+        self.telemetry_enabled = telemetry
+        
+        if telemetry:
+            asyncio.ensure_future(telemetry_extension.start())
 
     async def __fetch(self, op: str, data: dict):
         fetch = str(uuid4())
@@ -137,7 +161,15 @@ class DogeClient(Client):
 
             async def execute_command(command_name: str, ctx: Context, *args):
                 command = self.__commands.get(command_name.lower())
+            
                 if command:
+                    instance_id = f"{command_name}-{ctx.author.id}"
+                    invoked_at = time()
+                    if command[2] and instance_id in self.__command_cooldown_cache:
+                        passed = invoked_at - self.__command_cooldown_cache[instance_id]
+                        if passed < command[2]:
+                            return await execute_listener("on_cooldown_trigger", ctx, command_name, command[0], command[2] - passed)
+                    
                     arguments = []
                     params = {}
                     parameters = list(signature(command[0]).parameters.items())
@@ -148,21 +180,26 @@ class DogeClient(Client):
                     if parameters:
                         arguments.append(ctx)
                         parameters.pop(0)
+                        
                         for idx, (key, param) in enumerate(parameters):
                             if idx + 1 > len(args) and param.default != Parameter.empty:
-                                value = param.default
+                                params[key] = param.default
+                                continue
                             else:
                                 value = args[idx]
 
-                                if param.kind == param.KEYWORD_ONLY:
+                                if param.kind == param.KEYWORD_ONLY and len(args) - idx - 1 != 0:
                                     value = " ".join(args[idx::])
+                                    
+                            value = value.strip()
 
-                            if param.annotation and hasattr(param.annotation, "convert") and callable(param.annotation.convert):
+                            if value and param.annotation and hasattr(param.annotation, "convert") and callable(param.annotation.convert):
                                 value = await param.annotation.convert(ctx, param, value)
                             else:
                                 value = Convertor._handle_basic_types(param, value)
 
                             params[key] = value
+                        self.__command_cooldown_cache[instance_id] = invoked_at
                     try:
                         asyncio.ensure_future(command[0](*arguments, **params))
                     except TypeError:
@@ -198,27 +235,29 @@ class DogeClient(Client):
                             self.room = Room.from_dict(res["d"]["room"])
                             self.room.users = [self.user]
                         elif fetch == "get_user_profile":
+                            # TODO: USE THIS FOR GLOBAL CACHE
                             usr = User.from_dict(res["d"])
                             info(f"Dogehouse: Received user `{usr.id}`")
-                            self.room.users = [
-                                (user if user.id != usr.id else usr) for user in self.room.users]
+                            self.room.users = [(user if user.id != usr.id else usr) for user in self.room.users]
                             await execute_listener("on_user_fetch", usr)
                 elif op == "you-joined-as-speaker":
                     await execute_listener("on_room_join", True)
                 elif op == "join_room_done":
                     self.room = Room.from_dict(res["d"]["room"])
                     self.room.users.append(self.user)
-                    for user in self.room.users:
-                        if not isinstance(user, User):
-                            await self.__fetch("get_user_profile", dict(userId=user.id))
+                    await self.__send("get_current_room_users", {})
+                    # for user in self.room.users:
+                    #     if not isinstance(user, User):
+                    #         await self.__fetch("get_user_profile", dict(userId=user.id))
+                    
+                    # TODO: Check if joined as speaker 
                     await execute_listener("on_room_join", False)
                 elif op == "new_user_join_room":
                     user = User.from_dict(res["d"]["user"])
                     self.room.users.append(user)
                     await execute_listener("on_user_join", user)
                 elif op == "user_left_room":
-                    user = [user for user in self.room.users if user.id ==
-                            res["d"]["userId"]][0]
+                    user = [user for user in self.room.users if user.id == res["d"]["userId"]][0]
                     self.room.users.remove(user)
                     await execute_listener("on_user_leave", user)
                 elif op == "new_chat_msg":
@@ -231,8 +270,7 @@ class DogeClient(Client):
                     try:
                         async def handle_command(prefix: str):
                             if msg.content.startswith(prefix) and len(msg.content) > len(prefix) + 1:
-                                splitted = msg.content[len(
-                                    prefix)::].split(" ")
+                                splitted = msg.content[len(prefix)::].split(" ")
                                 await execute_command(splitted[0], Context(self, msg), *splitted[1::])
                                 return True
                             return False
@@ -244,7 +282,10 @@ class DogeClient(Client):
                             if await handle_command(prefix):
                                 break
                     except Exception as e:
-                        await execute_listener("on_error", e)
+                        if "on_error" not in self.__listeners:
+                            print_exc()
+                        else: 
+                            await execute_listener("on_error", e)
                 elif op == "message_deleted":
                     await execute_listener("on_message_delete", res["d"]["deleterId"], res["d"]["messageId"])
                 elif op == "speaker_added":
@@ -255,10 +296,15 @@ class DogeClient(Client):
                     await execute_listener("on_user_ban", res["d"]["userId"])
                 elif op == "hand_raised":
                     await execute_listener("on_speaker_request", res["d"]["userId"], res["d"]["roomId"])
+                elif op == "get_current_room_users_done":
+                    self.room.users = list(map(User.from_dict, res["d"]["users"]))
+                    await execute_listener("on_room_users_fetch")
 
         async def heartbeat():
             debug("Dogehouse: Starting heartbeat")
             while self.__active:
+                if self.telemetry_enabled:
+                    await telemetry_extension.transmit(self)
                 await self.__socket.send("ping")
                 await asyncio.sleep(heartbeatInterval)
 
@@ -450,7 +496,7 @@ class DogeClient(Client):
         Make a user in the room moderator.
 
         Args:
-            user (Union[User, BaseUser, UserPreview]): The user which should be promoted to room moderator.
+            user (Union[str, User, BaseUser, UserPreview]): The user which should be promoted to room moderator.
         """
         await self.__send("change_mod_status", dict(userId=str(user if isinstance(user, str) else user.id), value=True))
 
@@ -459,7 +505,7 @@ class DogeClient(Client):
         Remove a user their room moderator permissions.
 
         Args:
-            user (Union[User, BaseUser, UserPreview]): The user from which his permissions should be taken.
+            user (Union[str, User, BaseUser, UserPreview]): The user from which his permissions should be taken.
         """
         await self.__send("change_mod_status", dict(userId=str(user if isinstance(user, str) else user.id), value=False))
 
@@ -469,7 +515,7 @@ class DogeClient(Client):
         NOTE: This action is irreversable.
 
         Args:
-            user (Union[User, BaseUser, UserPreview]): The user which should be promoted to room admin.
+            user (Union[str, User, BaseUser, UserPreview]): The user which should be promoted to room admin.
         """
         await self.__send("change_room_creator", dict(userId=str(user if isinstance(user, str) else user.id)))
 
@@ -478,7 +524,7 @@ class DogeClient(Client):
         Force a user to be a listener.
 
         Args:
-            user (Union[User, BaseUser, UserPreview], optional): The user which should become a Listener. Defaults to the client.
+            user (Union[str, User, BaseUser, UserPreview], optional): The user which should become a Listener. Defaults to the client.
         """
         if not user:
             user = self.user
@@ -490,7 +536,7 @@ class DogeClient(Client):
         NOTE: This action can not be undone.
 
         Args:
-            user (Union[User, BaseUser, UserPreview]): The user from which their chat permissions should be taken.
+            user (Union[str, User, BaseUser, UserPreview]): The user from which their chat permissions should be taken.
         """
         await self.__send("ban_from_room_chat", dict(userId=str(user if isinstance(user, str) else user.id)))
 
@@ -499,7 +545,7 @@ class DogeClient(Client):
         Bans a user from a room.
 
         Args:
-            user (Union[User, BaseUser, UserPreview]): The user who should be banned.
+            user (Union[str, User, BaseUser, UserPreview]): The user who should be banned.
         """
         await self.__send("block_from_room", dict(userId=str(user if isinstance(user, str) else user.id)))
 
@@ -508,7 +554,7 @@ class DogeClient(Client):
         Unban a user from the room.
 
         Args:
-            user (Union[User, BaseUser, UserPreview]): The user who should be unbanned.
+            user (Union[str, User, BaseUser, UserPreview]): The user who should be unbanned.
         """
         await self.__send("unban_from_room", dict(userId=str((user if isinstance(user, str) else user.id)), fetch_id=uuid4()))
 
@@ -517,7 +563,7 @@ class DogeClient(Client):
         Accept a speaker request from a user.
 
         Args:
-            user (Union[User, BaseUser, UserPreview]): The user who will has to be accepted.
+            user (Union[str, User, BaseUser, UserPreview]): The user who will has to be accepted.
         """
         await self.__send("add_speaker", dict(userId=str(user if isinstance(user, str) else user.id)))
 
